"""Base Classes for Layers."""

from functools import reduce

import tensorflow as tf


class Layer:
    """Layer base class.

    This is an identity layer, and is primarily meant to be subclassed to
    construct more intersting layers.
    """

    def __call__(self, X):
        r"""Construct the subgraph for this layer.

        Parameters
        ----------
        X : Tensor
            the input to this layer

        Returns
        -------
        Net : Tensor
            the output of this layer
        KL : float, Tensor
            the regularizer/Kullback Leibler 'cost' of the parameters in this
            layer.

        """
        Net, KL = self._build(X)
        return Net, KL

    def _build(self, X):
        """Implement graph construction. Should be over-ridden."""
        return X, 0.0

    def __rshift__(self, other):
        """Implement layer composition, other(self(x))."""
        return LayerComposite(self, other)


class MultiLayer:
    """Base class for layers that take multiple inputs as kwargs.

    This is an Abstract class as there is no canonical identity for this
    layer (because it must do some kind of reduction).

    """

    def __call__(self, **kwargs):
        r"""Construct the subgraph for this layer.

        Parameters
        ----------
        **kwargs :
            the inputs to this layer (Tensors)

        Returns
        -------
        Net : Tensor
            the output of this layer
        KL : float, Tensor
            the regularizer/Kullback Leibler 'cost' of the parameters in this
            layer.

        """
        Net, KL = self._build(**kwargs)
        return Net, KL

    def _build(self, **kwargs):
        """Implement graph construction. Should be over-ridden."""
        raise NotImplementedError("Base class for MultiLayers only!")

    def __rshift__(self, other):
        """Implement multi-layer composition, other(self(x))."""
        return MultiLayerComposite(self, other)


class MultiLayerComposite(MultiLayer):
    """Composition of MultiLayers.

    Parameters
    ----------
    *layers :
        the layers to compose. First layer must be of type Multilayer,
        subsequent layers must be of type Layer.

    """

    def __init__(self, *layers):
        """Construct a new object."""
        self.stack = reduce(_stack2, layers)  # foldl

    def _build(self, **kwargs):
        """Stack the layers using function composition."""
        Net, KL = self.stack(**kwargs)
        return Net, KL


class LayerComposite(Layer):
    """Composition of Layers.

    Parameters
    ----------
    *layers :
        the layers to compose. All must be of type Layer.

    """

    def __init__(self, *layers):
        """Construct a new object."""
        self.stack = reduce(_stack2, layers)  # foldl

    def _build(self, X):
        """Stack the layers using function composition."""
        Net, KL = self.stack(X)
        return Net, KL


def stack(l, *layers):
    """Stack multiple Layers.

    This is a convenience function that acts as an alternative to the
    rshift operator implemented for Layers and Multilayers. It is syntatically
    more compact for stacking large numbers of layers or lists of layers.

    The type of stacking (Layer or Multilayer) is dispatched on the first
    argument.

    Parameters
    ----------
    l : Layer or MultiLayer
        The first layer to stack. The type of this layer determines the type
        of the output; MultiLayerComposite or LayerComposite.

    *layers :
        list of additional layers to stack. Must all be of type Layer,
        because function composition only works with the first function having
        multiple arguments.

    Returns
    -------
    result : MultiLayerComposite or LayerComposite
        A single layer that is the composition of the input layers.

    """
    if isinstance(l, MultiLayer):
        result = MultiLayerComposite(l, *layers)
    else:
        result = LayerComposite(l, *layers)
    return result


def _stack2(layer1, layer2):
    """Stack 2 functions, by composing w.r.t tensor, adding w.r.t losses."""
    def stackfunc(*args, **kwargs):
        result1, loss1 = layer1(*args, **kwargs)
        result, loss2 = layer2(result1)
        loss = tf.add(loss1, loss2)
        return result, loss
    return stackfunc



"""Network loss functions."""
import tensorflow as tf


def elbo(likelihood, Y, N, KL, like_weights=None):
    r"""Build the evidence lower bound loss for a neural net.

    Parameters
    ----------
    likelihood : tf.distributions.Distribution
        the likelihood object that takes neural network(s) as an input. The
        ``batch_shape`` of this object should be ``(n_samples, N, ...)``, where
        ``n_samples`` is the number of likelihood samples (defined by
        ab.InputLayer) and ``N`` is the number of observations (can be ``?`` if
        you are using a placeholder and mini-batching).
    Y : ndarray, Tensor
        the targets of shape ``(N, tasks)``.
    N : int, Tensor
        the total size of the dataset (i.e. number of observations).
    KL : float, Tensor
        the Kullback Leibler divergence between the posterior and prior
        parameters of the model (:math:`\text{KL}[q\|p]`).
    like_weights : callable, ndarray, Tensor
        weights to apply to each observation in the expected log likelihood.
        This should be an array of shape ``(N,)`` or can be called as
        ``like_weights(Y)`` and should return a ``(N,)`` array.

    Returns
    -------
    nelbo : Tensor
        the loss function of the Bayesian neural net (negative ELBO).

    """
    rank = len(likelihood.batch_shape)
    assert rank >= 2, "likelihood should be at least a rank 2 tensor! The " \
        "first dimension the network samples, the second the observations."

    # Batch amplification factor
    B = N / tf.to_float(likelihood.batch_shape_tensor()[1])

    # averaging over samples
    n_samples = tf.to_float(likelihood.batch_shape_tensor()[0])

    # Just mean over samps for expected log-likelihood
    ELL = _sum_likelihood(likelihood, Y, like_weights) / n_samples

    # negative ELBO is batch weighted ELL and KL
    nELBO = - B * ELL + KL

    return nELBO


def max_posterior(likelihood, Y, regulariser, like_weights=None,
                  observation_axis=-2):
    r"""Build maximum a-posteriori (MAP) loss for a neural net.

    Parameters
    ----------
    likelihood : tf.distributions.Distribution
        the likelihood object that takes neural network(s) as an input. One
        axis of this object's ``batch_shape`` should refer to ``N``, which is
        the number of observations (can be ``?`` if you are using a placeholder
        and mini-batching). The axis that indexes observations is set by the
        ``observation_axis`` setting. For data that is ``(N, tasks)`` or
        ``(..., N, tasks)``, this should automatically work.
    Y : ndarray, Tensor
        the targets of shape ``(N, tasks)``.
    regulariser : float, Tensor
        the regulariser on the parameters of the model to penalise model
        complexity.
    like_weights : callable, ndarray, Tensor
        weights to apply to each observation in the expected log likelihood.
        This should be an array of shape ``(N,)`` or can be called as
        ``like_weights(Y)`` and should return a ``(N,)`` array.
    observation_axis : int
        The axis that indexes the observations. This will assume the
        observations are on the *second last* axis, i.e. ``(..., N, D)``. This
        is used to calculate the mini-batch size.

    Returns
    -------
    map : Tensor
        the loss function of the MAP neural net.

    """
    # Get the batch size to average the likelihood over
    M = tf.to_float(likelihood.batch_shape_tensor()[observation_axis])

    # Average likelihood for batch
    AVLL = _sum_likelihood(likelihood, Y, like_weights) / M

    # MAP objective
    MAP = - AVLL + regulariser

    return MAP


#
# Private module utilities
#

def _sum_likelihood(likelihood, Y, like_weights):
    """Sum the log-likelihood of the Y's under the model."""
    log_prob = likelihood.log_prob(Y)
    if callable(like_weights):
        log_prob *= like_weights(Y)
    elif like_weights is not None:
        log_prob *= like_weights

    sumlike = tf.reduce_sum(log_prob)

    return sumlike



"""Dataset generation and loading utilitites."""
import os

import numpy as np
from scipy.io import loadmat
from sklearn.gaussian_process.kernels import RBF
from sklearn.datasets.base import Bunch
from six.moves import urllib

from aboleth.random import seedgen

DEFAULT_DATA_PATH = os.path.join(os.path.dirname(__file__), '../data/')


def gp_draws(ntrain, ntest, kern=RBF(length_scale=0.5), noise=0.1, xmin=-10,
             xmax=10):
    r"""Generate a random (noisy) draw from a Gaussian Process.

    Parameters
    ----------
    ntrain : int
        number of training points to generate
    ntest : int
        number of testing points to generate
    kern : scikit.gaussian_process.kernels
        kernel to generate data from
    noise : float
        Gaussian noise (standard deviation) to add to GP draws
    xmin : float
        minimum extent of inputs, X
    xmax : float
        maximum extent of inputs, X

    Returns
    -------
    Xtrain : ndarray
        of shape (ntrain, 1) of training inputs
    Ytrain : ndarray
        of shape (ntrain, 1) of training targets
    Xtest : ndarray
        of shape (ntrain, 1) of testing inputs
    Ytest : ndarray
        of shape (ntrain, 1) of testing targets
    """
    randgen = np.random.RandomState(next(seedgen))

    Xtrain = randgen.rand(ntrain)[:, np.newaxis] * (xmin - xmax) - xmin
    Xtest = np.linspace(xmin, xmax, ntest)[:, np.newaxis]
    Xcat = np.vstack((Xtrain, Xtest))

    K = kern(Xcat, Xcat)
    U, S, V = np.linalg.svd(K)
    L = U.dot(np.diag(np.sqrt(S))).dot(V)
    f = randgen.randn(ntrain + ntest).dot(L)

    Ytrain = f[0:ntrain] + randgen.randn(ntrain) * noise
    ftest = f[ntrain:]

    Xtrain = Xtrain.astype(np.float32)
    Ytrain = Ytrain[:, np.newaxis].astype(np.float32)
    Xtest = Xtest.astype(np.float32)
    ftest = ftest[:, np.newaxis].astype(np.float32)
    return Xtrain, Ytrain, Xtest, ftest


def fetch_gpml_sarcos_data():
    r"""Fetch the SARCOS dataset.

    Fetch the SARCOS dataset from the internet and parse appropriately into
    python arrays

    Returns
    -------
    data : sklearn.datasets.Bunch
        Bunch object that contrains the dataset

    Examples
    --------
    >>> gpml_sarcos = fetch_gpml_sarcos_data()

    >>> gpml_sarcos.train.data.shape
    (44484, 21)

    >>> gpml_sarcos.train.targets.shape
    (44484,)

    >>> gpml_sarcos.train.targets.round(2) # doctest: +ELLIPSIS
    array([ 50.29,  44.1 ,  37.35, ...,  22.7 ,  17.13,   6.52])

    >>> gpml_sarcos.test.data.shape
    (4449, 21)

    >>> gpml_sarcos.test.targets.shape
    (4449,)
    """
    train_src_url = "http://www.gaussianprocess.org/gpml/data/sarcos_inv.mat"
    test_src_url = ("http://www.gaussianprocess.org/gpml/data/sarcos_inv_test"
                    ".mat")

    train_filename = os.path.join(DEFAULT_DATA_PATH, 'sarcos_inv.mat')
    test_filename = os.path.join(DEFAULT_DATA_PATH, 'sarcos_inv_test.mat')

    if not os.path.exists(train_filename):
        urllib.request.urlretrieve(train_src_url, train_filename)

    if not os.path.exists(test_filename):
        urllib.request.urlretrieve(test_src_url, test_filename)

    train_data = loadmat(train_filename).get('sarcos_inv')
    test_data = loadmat(test_filename).get('sarcos_inv_test')

    train_bunch = Bunch(data=train_data[:, :21],
                        targets=train_data[:, 21])

    test_bunch = Bunch(data=test_data[:, :21],
                       targets=test_data[:, 21])

    return Bunch(train=train_bunch, test=test_bunch)



"""Helper functions for model parameter distributions."""
import numpy as np
import tensorflow as tf

from tensorflow.contrib.distributions import MultivariateNormalTriL

from aboleth.util import pos
from aboleth.random import seedgen


#
# Streamlined interfaces for initialising the priors and posteriors
#

def norm_prior(dim, std):
    """Initialise a prior (zero mean, isotropic) Normal distribution.

    Parameters
    ----------
    dim : tuple or list
        the dimension of this distribution.
    std : float
        the prior standard deviation of this distribution.

    Returns
    -------
    P : tf.distributions.Normal
        the initialised prior Normal object.

    Note
    ----
    This will make a tf.Variable on the variance of the prior that is
    initialised with ``std``.

    """
    mu = tf.zeros(dim)
    std = pos(tf.Variable(std, name="W_mu_p"))
    P = tf.distributions.Normal(loc=mu, scale=std)
    return P


def norm_posterior(dim, std0):
    """Initialise a posterior (diagonal) Normal distribution.

    Parameters
    ----------
    dim : tuple or list
        the dimension of this distribution.
    std0 : float
        the initial (unoptimized) standard deviation of this distribution.

    Returns
    -------
    Q : tf.distributions.Normal
        the initialised posterior Normal object.

    Note
    ----
    This will make tf.Variables on the randomly initialised mean and standard
    deviation of the posterior. The initialisation of the mean is from a Normal
    with zero mean, and ``std0`` standard deviation, and the initialisation of
    the standard deviation is from a gamma distribution with an alpha of
    ``std0`` and a beta of 1.

    """
    mu_0 = tf.random_normal(dim, stddev=std0, seed=next(seedgen))
    mu = tf.Variable(mu_0, name="W_mu_q")

    std_0 = tf.random_gamma(alpha=std0, shape=dim, seed=next(seedgen))
    std = pos(tf.Variable(std_0, name="W_std_q"))

    Q = tf.distributions.Normal(loc=mu, scale=std)
    return Q


def gaus_posterior(dim, std0):
    """Initialise a posterior Gaussian distribution with a diagonal covariance.

    Even though this is initialised with a diagonal covariance, a full
    covariance will be learned, using a lower triangular Cholesky
    parameterisation.

    Parameters
    ----------
    dim : tuple or list
        the dimension of this distribution.
    std0 : float
        the initial (unoptimized) diagonal standard deviation of this
        distribution.

    Returns
    -------
    Q : tf.contrib.distributions.MultivariateNormalTriL
        the initialised posterior Gaussian object.

    Note
    ----
    This will make tf.Variables on the randomly initialised mean and covariance
    of the posterior. The initialisation of the mean is from a Normal with zero
    mean, and ``std0`` standard deviation, and the initialisation of the (lower
    triangular of the) covariance is from a gamma distribution with an alpha of
    ``std0`` and a beta of 1.

    """
    O, I = dim

    # Optimize only values in lower triangular
    u, v = np.tril_indices(I)
    indices = (u * I + v)[:, np.newaxis]
    l0 = np.tile(np.eye(I), [O, 1, 1])[:, u, v].T
    l0 = l0 * tf.random_gamma(alpha=std0, shape=l0.shape, seed=next(seedgen))
    l = tf.Variable(l0, name="W_cov_q")
    Lt = tf.transpose(tf.scatter_nd(indices, l, shape=(I * I, O)))
    L = tf.reshape(Lt, (O, I, I))

    mu_0 = tf.random_normal((O, I), stddev=std0, seed=next(seedgen))
    mu = tf.Variable(mu_0, name="W_mu_q")
    Q = MultivariateNormalTriL(mu, L)
    return Q


#
# KL divergence calculations
#

def kl_sum(q, p):
    r"""Compute the total KL between (potentially) many distributions.

    I.e. :math:`\sum_i \text{KL}[q_i || p_i]`

    Parameters
    ----------
    q : tf.distributions.Distribution
        A tensorflow Distribution object
    p : tf.distributions.Distribution
        A tensorflow Distribution object

    Returns
    -------
    kl : Tensor
        the result of the sum of the KL divergences of the ``q`` and ``p``
        distibutions.
    """
    kl = tf.reduce_sum(tf.distributions.kl_divergence(q, p))
    return kl


@tf.distributions.RegisterKL(MultivariateNormalTriL, tf.distributions.Normal)
def _kl_gaussian_normal(q, p, name=None):
    """Gaussian-Normal Kullback Leibler divergence calculation.

    Parameters
    ----------
    q : tf.contrib.distributions.MultivariateNormalTriL
        the approximating 'q' distribution(s).
    p : tf.distributions.Normal
        the prior 'p' distribution(s), ``p.scale`` should be a *scalar* value!
    name : str
        name to give the resulting KL divergence Tensor

    Returns
    -------
    KL : Tensor
        the result of KL[q||p].

    """
    assert len(p.scale.shape) == 0, "This KL divergence is only implemented " \
        "for Normal distributions that share a scale parameter for p"
    D = tf.to_float(q.event_shape_tensor())
    n = tf.to_float(q.batch_shape_tensor())
    p_var = p.scale**2
    L = q.scale.to_dense()
    tr = tf.reduce_sum(L * L) / p_var
    # tr = tf.reduce_sum(tf.trace(q.covariance())) / p_var  # Above is faster
    dist = tf.reduce_sum((p.mean() - q.mean())**2) / p_var
    # logdet = n * D * tf.log(p_var) \
    #     - 2 * tf.reduce_sum(q.scale.log_abs_determinant())  # Numerical issue
    logdet = n * D * tf.log(p_var) - _chollogdet(L)
    KL = 0.5 * (tr + dist + logdet - n * D)
    if name:
        KL = tf.identity(KL, name=name)
    return KL


#
# Private module stuff
#

def _chollogdet(L):
    """Log det of a cholesky, where L is (..., D, D)."""
    l = pos(tf.matrix_diag_part(L))  # keep > 0, and no vanashing gradient
    logdet = 2. * tf.reduce_sum(tf.log(l))
    return logdet



"""Test the layers module."""
import pytest
import numpy as np
import tensorflow as tf
import aboleth as ab

from aboleth.distributions import norm_prior, gaus_posterior
from aboleth.layers import SampleLayer


D = 10
DIM = (10, 2)
EDIM = (10, 5)


def test_net_outputs(make_graph):
    """Test for expected output dimensions from a net."""
    x, y, N, X_, Y_, N_, layers = make_graph
    Net, kl = layers(X=X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()

        P = Net.eval(feed_dict={X_: x, Y_: y, N_: float(N)})

        assert P.shape == (10, N, 1)


def test_input(make_data):
    """Test the input layer."""
    x, _, X = make_data
    s = ab.InputLayer(name='myname')

    F, KL = s(myname=x)
    tc = tf.test.TestCase()
    with tc.test_session():
        f = F.eval()
        assert KL == 0.0
        assert np.array_equal(f, x)


def test_input_sample(make_data):
    """Test the input and tiling layer."""
    x, _, X = make_data
    s = ab.InputLayer(name='myname', n_samples=3)

    F, KL = s(myname=x)
    tc = tf.test.TestCase()
    with tc.test_session():
        f = F.eval()
        X_array = X.eval()
        assert KL == 0.0
        assert np.array_equal(f, X_array)
        for i in range(3):
            assert np.array_equal(f[i], x)


def test_activation(make_data):
    """Test nonlinear activation layer."""
    x, _, X = make_data
    act = ab.Activation(tf.tanh)

    tc = tf.test.TestCase()
    with tc.test_session():
        F, KL = act(X)

        assert np.all(np.tanh(X.eval()) == F.eval())
        assert KL == 0


def test_dropout(make_data):
    """Test dropout layer."""
    x, _, X = make_data
    drop = ab.DropOut(0.5)

    F, KL = drop(X)

    tc = tf.test.TestCase()
    with tc.test_session():
        f = F.eval()
        prop_zero = np.sum(f == 0) / np.prod(f.shape)

        assert f.shape == X.eval().shape
        assert (prop_zero > 0.4) and (prop_zero < 0.6)
        assert KL == 0


def test_max_pooling2d(make_image_data):
    """Test dropout layer."""
    x, _, X = make_image_data

    # downsample by 2x
    max_pool = ab.MaxPool2D(pool_size=(2, 2),
                            strides=(2, 2))

    F, KL = max_pool(X)

    tc = tf.test.TestCase()
    with tc.test_session():
        f = F.eval()

        # test equivalence of first window across batches
        assert np.all(np.max(x[:, :2, :2, :], axis=(1, 2)) == f[0, :, 0, 0, :])

        # n_samples and batch size remain unchanged
        assert f.shape[:2] == X.eval().shape[:2]

        # downsampled by 2x
        assert 2 * f.shape[2] == X.eval().shape[2]
        assert 2 * f.shape[3] == X.eval().shape[3]

        # number of channels remain unchanged
        assert f.shape[-1] == X.eval().shape[-1]

        assert KL == 0


def test_reshape(make_image_data):
    """Test dropout layer."""
    x, _, X = make_image_data

    # reshape. useful for feeding output of conv layer into dense layer
    reshape = ab.Reshape(target_shape=(28*28*3,))

    F, KL = reshape(X)

    tc = tf.test.TestCase()
    with tc.test_session():
        f = F.eval()

        assert f.shape[-1] == np.prod(X.eval().shape[2:])
        assert np.all(f == np.reshape(X.eval(), (5, 100, 28*28*3)))

        assert KL == 0


def test_arc_cosine(make_data):
    """Test the random Arc Cosine kernel."""
    S = 3
    x, _, _ = make_data
    x_, X_ = _make_placeholders(x, S)

    F, KL = ab.RandomArcCosine(n_features=10)(X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        f = F.eval(feed_dict={x_: x})

        assert f.shape == (3, x.shape[0], 10)
        assert KL == 0


def test_dense_embeddings(make_categories):
    """Test the embedding layer."""
    x, K = make_categories
    N = len(x)
    S = 3
    x_, X_ = _make_placeholders(x, S, tf.int32)
    output, KL = ab.EmbedVariational(output_dim=D, n_categories=K)(X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()
        kl = KL.eval()

        assert np.isscalar(kl)
        assert kl > 0

        Phi = output.eval(feed_dict={x_: x})

        assert Phi.shape == (S, N, D)


@pytest.mark.parametrize('dense', [ab.DenseMAP, ab.DenseVariational])
def test_dense_outputs(dense, make_data):
    """Make sure the dense layers output expected dimensions."""
    x, _, _ = make_data
    S = 3

    x_, X_ = _make_placeholders(x, S)
    N = x.shape[0]

    Phi, KL = dense(output_dim=D)(X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()
        P = Phi.eval(feed_dict={x_: x})
        assert P.shape == (S, N, D)
        assert P.dtype == np.float32
        assert np.isscalar(KL.eval(feed_dict={x_: x}))


@pytest.mark.parametrize('layer_args', [
    (SampleLayer, ()),
    (ab.DenseMAP, (D,)),
    (ab.DenseVariational, (D,)),
    (ab.EmbedVariational, (2, D)),
    (ab.RandomFourier, (2, ab.RBF())),
    (ab.RandomArcCosine, (2,)),
])
def test_sample_layer_input_exception(layer_args, make_data):
    """Make sure sample layers fail when the don't get a rank 3 tensor."""
    x, _, _ = make_data
    layer, args = layer_args
    with pytest.raises(AssertionError):
        layer(*args)(x)


@pytest.mark.parametrize('kernels', [
    (ab.RBF, {}),
    (ab.RBFVariational, {}),
    (ab.Matern, {'p': 1}),
    (ab.Matern, {'p': 2})
])
def test_fourier_features(kernels, make_data):
    """Test random fourier kernels approximations."""
    D = 100
    S = 3
    kern, p = kernels
    k = kern(D, **p)

    x, _, _ = make_data
    x_, X_ = _make_placeholders(x, S)
    N = x.shape[0]

    Phi, KL = ab.RandomFourier(D, k)(X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()
        P = Phi.eval(feed_dict={x_: x})
        for i in range(P.shape[0]):
            p = P[i]
            assert p.shape == (N, 2 * D)
            # Check behaving properly with k(x, x) ~ 1.0
            assert np.allclose((p**2).sum(axis=1), np.ones(N))

        # Make sure we get a valid KL
        kl = KL.eval() if isinstance(KL, tf.Tensor) else KL
        assert kl >= 0


@pytest.mark.parametrize('dists', [
    {'prior_W': norm_prior(DIM, 1.), 'prior_b': norm_prior((D,), 1.)},
    {'post_W': norm_prior(DIM, 1.), 'post_b': norm_prior((D,), 1.)},
    {'prior_W': norm_prior(DIM, 1.), 'post_W': norm_prior(DIM, 1.)},
    {'prior_W': norm_prior(DIM, 1.), 'post_W': gaus_posterior(DIM, 1.)},
    {'prior_W': gaus_posterior(DIM, 1.), 'post_W': gaus_posterior(DIM, 1.)},
])
def test_dense_distribution(dists, make_data):
    """Test initialising dense variational layers with distributions."""
    x, _, _ = make_data
    S = 3

    x_, X_ = _make_placeholders(x, S)
    N = x.shape[0]

    Phi, KL = ab.DenseVariational(output_dim=D, **dists)(X_)
    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()
        P = Phi.eval(feed_dict={x_: x})
        assert P.shape == (S, N, D)
        assert KL.eval() >= 0.


@pytest.mark.parametrize('dists', [
    {'prior_W': norm_prior(EDIM, 1.), 'post_W': norm_prior(EDIM, 1.)},
    {'prior_W': norm_prior(EDIM, 1.), 'post_W': gaus_posterior(EDIM, 1.)},
    {'prior_W': gaus_posterior(EDIM, 1.), 'post_W': gaus_posterior(EDIM, 1.)},
])
def test_embeddings_distribution(dists, make_categories):
    """Test initialising embedding variational layers with distributions."""
    x, K = make_categories
    N = len(x)
    S = 3
    x_, X_ = _make_placeholders(x, S, tf.int32)
    output, KL = ab.EmbedVariational(output_dim=D, n_categories=K, **dists)(X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()
        Phi = output.eval(feed_dict={x_: x})
        assert Phi.shape == (S, N, D)
        assert KL.eval() >= 0.


def _make_placeholders(x, S, xtype=tf.float32):
    x_ = tf.placeholder(xtype, x.shape)
    X_ = tf.tile(tf.expand_dims(x_, 0), [S, 1, 1])
    return x_, X_



#!/usr/bin/env python

from setuptools import setup, find_packages

exec(open('aboleth/version.py').read())
readme = open('README.rst').read()

setup(
    name='aboleth',
    version=__version__,
    description='Bayesian supervised deep learning with TensorFlow',
    long_description=readme,
    author='Daniel Steinberg, Lachlan McCalman',
    author_email='daniel.steinberg@data61.csiro.au',
    url='https://github.com/data61/aboleth',
    packages=find_packages(),
    package_dir={'aboleth': 'aboleth'},
    include_package_data=True,
    install_requires=[
        'numpy>=1.12.0',
        'scipy>=0.18.1',
        'tensorflow>=1.3.0',
        'six>=1.10.0'
    ],
    extras_require={
        'dev': [
            'sphinx>=1.4.8',
            'pytest>=3.1.0',
            'pytest-mock>=1.6.0',
            'pytest-cov>=2.5.1',
            'pytest-flake8>=0.8.1',
            'flake8-docstrings>=1.1.0',
        ],
        'demos': [
            'bokeh>=0.12.4',
            'pandas>=0.20.3',
            'scikit-learn>=0.18.1',
        ]
    },
    license="Apache 2.0",
    zip_safe=False,
    keywords='aboleth',
    classifiers=[
        'Development Status :: 4 - Beta',
        "Operating System :: POSIX",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis"
    ],
)
